# -*- coding: utf-8 -*-
# Copyright 2016 Camptocamp SA
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html)

from __future__ import print_function

import argparse
import code
import importlib
import logging
import signal

from contextlib import contextmanager

from anthem import __version__ as anthem_version
import openerp
from openerp.api import Environment

from .output import LogIndent


def main():
    parser = argparse.ArgumentParser(
        description='Anthem: make your Odoo scripts sing!')
    parser.add_argument(
        'target',
        help='the target to run in the format path.to.module::function. This '
        'will import path.to.module and run function'
    )
    parser.add_argument(
        '-i', '--interactive',
        action='store_true',
        help='after playing the song, open an interactive console with access '
        'to ctx (the the Anthem context), much like running python -i'
    )
    parser.add_argument(
        '-q', '--quiet',
        action='store_true',
        help='Quiet the logs, which is sad for the songs.',
    )
    parser.add_argument(
        'odoo-args',
        nargs=argparse.REMAINDER,
        help='command line arguments to be passed on to Odoo, like -c for a '
        'configuration file and -d for the database name. Those must be at '
        'the end of the arguments'
    )
    args = parser.parse_args()
    odoo_args = vars(args)['odoo-args']
    options = Options(interactive=args.interactive, quiet=args.quiet)
    run(odoo_args, args.target, options)


def banner():
    b = (
        'Welcome to the anthem version {} interactive console!\n'
        '\n'
        'You can use an anthem context as ctx. (ctx.env is an Odoo\n'
        'environment). If you terminate your session with Ctrl-D,\n'
        'the database cursor ctx.env.cr will be committed.\n'
        'Otherwise if you exit() or you raise an exception, the cursor will\n'
        'be closed without a commit.\n'
    ).format(anthem_version)
    return b


class Options(object):

    def __init__(self, interactive=False, quiet=False, test_mode=False):
        self.interactive = interactive
        self.quiet = quiet
        self.test_mode = test_mode


def run(odoo_args, target, options):
    with Context(odoo_args, options) as ctx:
        mod_name, func_name = target.split('::')
        module = importlib.import_module(mod_name)
        func = getattr(module, func_name)
        func(ctx)
        if options.interactive:
            console = code.InteractiveConsole(locals={'ctx': ctx})
            import rlcompleter  # noqa
            import readline
            readline.parse_and_bind("tab: complete")
            console.interact(banner=banner())
        ctx.env.cr.commit()


class Context(object):
    def __init__(self, odoo_args, options):
        self.env = self._build_odoo_env(odoo_args)
        self.options = options
        self._log = LogIndent()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.options.test_mode:
            self.env.cr.rollback()
        self.env.cr.close()

    def _build_odoo_env(self, odoo_args):
        openerp.tools.config.parse_config(odoo_args)
        dbname = openerp.tools.config['db_name']
        if not dbname:
            argparse.ArgumentParser().error(
                "please provide a database name though Odoo options (either "
                "-d or an Odoo configuration file)"
            )
        logging.getLogger('openerp').setLevel(logging.ERROR)
        openerp.service.server.start(preload=[], stop=True)

        # openerp.service.server.start() modifies the SIGINT signal by its own
        # one which in fact prevents us to stop anthem with Ctrl-c.
        # Restore the default one.
        signal.signal(signal.SIGINT, signal.default_int_handler)

        registry = openerp.modules.registry.RegistryManager.get(dbname)
        cr = registry.cursor()
        uid = openerp.SUPERUSER_ID
        Environment.reset()
        context = Environment(cr, uid, {})['res.users'].context_get()
        return Environment(cr, uid, context)

    def log_line(self, message):
        self._log.print_indent(message)

    @contextmanager
    def log(self, name, timing=True):
        if self.options.quiet:
            yield
        else:
            with self._log.display(name, timing=timing):
                yield



# -*- coding: utf-8 -*-

"""
Lyrics are isolated pieces of code that you can use and reuse in
your own songs.

They can be structured as you desire, from Haikus to Odes as long as the songs
you write reach the zen of Python and execute harmoniously.

Examples of lyrics are:

 * load a csv, xml or yaml file
 * add a xmlid on a record
 * upsert a record

"""



# -*- coding: utf-8 -*-
# Copyright 2016 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import csv

from ..exceptions import AnthemError


def load_csv(ctx, model_name, path, dialect='excel', **fmtparams):
    """ Load a CSV from a filename

    Usage example::

      from pkg_resources import Requirement, resource_string

      req = Requirement.parse('my-project')
      load_csv(ctx, 'res.users',
               resource_string(req, 'data/users.csv'),
               delimiter=',')

    """
    with open(path, 'rb') as data:
        load_csv_stream(ctx, model_name, data, dialect=dialect, **fmtparams)


def load_csv_stream(ctx, model_name, data, dialect='excel', **fmtparams):
    """ Load a CSV from a stream

    Usage example::

      from pkg_resources import Requirement, resource_stream

      req = Requirement.parse('my-project')
      load_csv_stream(ctx, 'res.users',
                      resource_stream(req, 'data/users.csv'),
                      delimiter=',')

    """
    data = csv.reader(data, dialect=dialect, **fmtparams)
    head = data.next()
    values = list(data)
    if values:
        result = ctx.env[model_name].load(head, values)
        ids = result['ids']
        if not ids:
            messages = u'\n'.join(
                u'- %s' % msg for msg in result['messages']
            )
            ctx.log_line(u"Failed to load CSV "
                         u"in '%s'. Details:\n%s" %
                         (model_name, messages))
            raise AnthemError(u'Could not import CSV. See the logs')
        else:
            ctx.log_line(u"Imported %d records in '%s'" %
                         (len(ids), model_name))



# -*- coding: utf-8 -*-
# Copyright 2016 Camptocamp SA
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html)

__version__ = "0.4.0"

# publish the decorator so we can use 'anthem.log'
from output import log  # noqa



# -*- coding: utf-8 -*-
# Copyright 2016 Camptocamp SA
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html)


def add_xmlid(ctx, record, xmlid, noupdate=False):
    """ Add a XMLID on an existing record """
    if '.' in xmlid:
        module, name = xmlid.split('.')
    else:
        module = ''
        name = xmlid
    return ctx.env['ir.model.data'].create({
        'name': name,
        'module': module,
        'model': record._name,
        'res_id': record.id,
        'noupdate': noupdate,
    })


def create_or_update(ctx, model, xmlid, values):
    """ Create or update a record matching xmlid with values """
    record = ctx.env.ref(xmlid, raise_if_not_found=False)
    if record:
        record.update(values)
    else:
        record = ctx.env[model].create(values)
        add_xmlid(ctx, record, xmlid)
    return record



# -*- coding: utf-8 -*-

import re

from setuptools import setup, find_packages

with open('anthem/__init__.py') as f:
    version = re.search('^__version__\s*=\s*"(.*)"', f.read(), re.M).group(1)
with open('README.rst') as f:
    readme = f.read()
with open('HISTORY.rst') as f:
    history = f.read()

setup(
    name='anthem',
    version=version,
    description='Make your Odoo scripts sing.',
    long_description=readme + '\n\n' + history,
    author='Camptocamp',
    author_email='info@camptocamp.com',
    url='https://github.com/camptocamp/anthem',
    license='LGPLv3+',
    packages=find_packages(exclude=('tests', 'docs')),
    entry_points={
        'console_scripts': ['anthem = anthem.cli:main']
    },
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ),
)



# -*- coding: utf-8 -*-
# Copyright 2016 Camptocamp SA
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl-3.0.en.html)

from __future__ import print_function

import functools
import sys
import time

from contextlib import contextmanager


def safe_print(ustring, errors='replace', **kwargs):
    """ Safely print a unicode string """
    encoding = sys.stdout.encoding or 'utf-8'
    bytestr = ustring.encode(encoding, errors=errors)
    print(bytestr, **kwargs)


class LogIndent(object):

    def __init__(self):
        self.level = 0

    @contextmanager
    def display(self, name, timing=True):
        self.print_indent(u'{}...'.format(name))
        self.level += 1
        start = time.time()
        try:
            yield
        except:
            self.level -= 1
            self.print_indent(u'{}: error'.format(name))
            raise
        end = time.time()
        self.level -= 1
        if timing:
            self.print_indent(u"{}: {:.3f}s".format(name, end - start))

    def print_indent(self, message):
        safe_print(u"{}{}".format(u"    " * self.level, message))


def log(func=None, name=None, timing=True):
    """ Decorator to show a description of the running function

    By default, it outputs the first line of the docstring.
    If the docstring is empty, it displays the name of the function.
    Alternatively, if a ``name`` is specified, it will display that only.

    It can be called as ``@log`` or as ``@log(name='abc, timing=True)``.

    """
    # support to be called as @log or as @log(name='')
    if func is None:
        return functools.partial(log, name=name, timing=timing)

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        assert len(args) > 0 and hasattr(args[0], 'log'), \
            "The first argument of the decorated function must be a Context"
        ctx = args[0]
        message = name
        if message is None:
            if func.__doc__:
                message = func.__doc__.splitlines()[0].strip()
        if message is None:
            message = func.func_name
        with ctx.log(message, timing=timing):
            return func(*args, **kwargs)
    return decorated
